"""Support for discovering myStrom devices."""

import asyncio
import logging
from typing import List, Optional

from .device_types import DEVICE_MAPPING_NUMERIC

_LOGGER = logging.getLogger(__name__)


class DiscoveredDevice(object):
    """Representation of discovered device."""

    mac: str
    type: int
    is_child: bool
    mystrom_registered: bool
    mystrom_online: bool
    restarted: bool

    @staticmethod
    def create_from_announce_msg(raw_addr, announce_msg):
        """Create announce message."""
        _LOGGER.debug("Received announce message '%s' from %s ", announce_msg, raw_addr)
        if len(announce_msg) != 8:
            raise RuntimeError("Unexpected announcement, '%s'" % announce_msg)

        device = DiscoveredDevice(host=raw_addr[0], mac=announce_msg[0:6].hex(":"))
        device.type = announce_msg[6]

        if device.type == "102":
            device.hardware = DEVICE_MAPPING_NUMERIC[int(announce_msg[6])]
        else:
            device.hardware = "non_mystrom"
        status = announce_msg[7]

        # Parse status field
        device.is_child = status & 1 != 0
        device.mystrom_registered = status & 2 != 0
        device.mystrom_online = status & 4 != 0
        device.restarted = status & 8 != 0
        return device

    def __init__(self, host, mac):
        """Initialize the discovery."""
        self.host = host
        self.mac = mac


class DeviceRegistry(object):
    """Representation of the device registry."""

    def __init__(self):
        """Initialize the device registry."""
        self.devices_by_mac = {}

    def register(self, device):
        """Register a device."""
        self.devices_by_mac[device.mac] = device

    def devices(self):
        """Get all present devices."""
        return list(self.devices_by_mac.values())


class DiscoveryProtocol(asyncio.DatagramProtocol):
    """Representation of the discovery protocol."""

    def __init__(self, registry: DeviceRegistry):
        """ "Initialize the discovery protocol."""
        super().__init__()
        self.registry = registry

    def connection_made(self, transport):
        """Create an UDP listener."""
        _LOGGER.debug("Starting up UDP listener")
        self.transport = transport

    def datagram_received(self, data, addr):
        """Handle a datagram."""
        device = DiscoveredDevice.create_from_announce_msg(addr, data)
        self.registry.register(device)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Stop if connection is lost."""
        _LOGGER.debug("Shutting down UDP listener")
        super().connection_lost(exc)


async def discover_devices(timeout: int = 7) -> List[DiscoveredDevice]:
    """Discover local myStrom devices.

    Some myStrom devices report their presence every ~5 seconds in an UDP
    broadcast to port 7979.
    """
    registry = DeviceRegistry()
    loop = asyncio.get_event_loop()
    (transport, protocol) = await loop.create_datagram_endpoint(
        lambda: DiscoveryProtocol(registry), local_addr=("0.0.0.0", 7979)
    )
    # Server runs in the background, meanwhile wait until timeout expires
    await asyncio.sleep(timeout)
    # Shutdown server
    transport.close()

    devices = registry.devices()
    for device in devices:
        _LOGGER.debug(
            "Discovered myStrom device %s (%s) (MAC addresse: %s)",
            device.host,
            device.type,
            device.mac,
        )
    return devices



"""Tests for myStrom bulbs."""

import pytest

import pymystrom.bulb as bulb_module
from pymystrom.bulb import MyStromBulb


@pytest.mark.asyncio
async def test_set_on(monkeypatch):
    """Test MyStromBulb.set_on sends correct request and returns response."""
    mac = "AA:BB:CC:DD:EE:FF"
    called = {}

    async def _fake_request(
        self, uri, method="POST", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to capture call parameters."""
        called.update({"uri": uri, "method": method, "data": data, "token": token})
        return {"result": "ok"}

    monkeypatch.setattr(bulb_module, "request", _fake_request)
    bulb = MyStromBulb("127.0.0.1", mac)
    resp = await bulb.set_on()
    assert called["method"] == "POST"
    assert called["data"] == {"action": "on"}
    assert resp == {"result": "ok"}


@pytest.mark.asyncio
async def test_set_off(monkeypatch):
    """Test MyStromBulb.set_off sends correct request and returns response."""
    mac = "AA:BB:CC:DD:EE:FF"
    called = {}

    async def _fake_request(
        self, uri, method="POST", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to capture call parameters."""
        called.update({"uri": uri, "method": method, "data": data, "token": token})
        return {"result": "ok"}

    monkeypatch.setattr(bulb_module, "request", _fake_request)
    bulb = MyStromBulb("127.0.0.1", mac)
    resp = await bulb.set_off()
    assert called["method"] == "POST"
    assert called["data"] == {"action": "off"}
    assert resp == {"result": "ok"}


@pytest.mark.asyncio
async def test_set_color_hex(monkeypatch):
    """Test MyStromBulb.set_color_hex sends correct request and returns response."""
    mac = "AA:BB:CC:DD:EE:FF"
    called = {}

    async def _fake_request(
        self, uri, method="POST", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to capture call parameters."""
        called.update({"uri": uri, "method": method, "data": data, "token": token})
        return {"result": "ok"}

    monkeypatch.setattr(bulb_module, "request", _fake_request)
    bulb = MyStromBulb("127.0.0.1", mac)
    resp = await bulb.set_color_hex("FF000000")
    assert called["data"]["color"] == "FF000000"
    assert resp == {"result": "ok"}


@pytest.mark.asyncio
async def test_set_color_hsv(monkeypatch):
    """Test MyStromBulb.set_color_hsv sends correct request and returns response."""

    mac = "AA:BB:CC:DD:EE:FF"
    called = {}

    async def _fake_request(
        self, uri, method="POST", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to capture call parameters."""
        called.update({"uri": uri, "method": method, "data": data, "token": token})
        return {"result": "ok"}

    monkeypatch.setattr(bulb_module, "request", _fake_request)
    bulb = MyStromBulb("127.0.0.1", mac)
    resp = await bulb.set_color_hsv(1, 2, 3)
    assert "color=1;2;3" in called["data"]
    assert resp == {"result": "ok"}


@pytest.mark.asyncio
async def test_set_white(monkeypatch):
    """Test MyStromBulb.set_white sends correct request and returns response."""

    mac = "AA:BB:CC:DD:EE:FF"
    called = {"calls": []}

    async def _fake_request(
        self, uri, method="POST", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to capture call parameters."""
        called["calls"].append(data)
        return {"result": "ok"}

    monkeypatch.setattr(bulb_module, "request", _fake_request)
    bulb = MyStromBulb("127.0.0.1", mac)
    await bulb.set_white()
    # Should call set_color_hsv with (0, 0, 100)
    assert any("color=0;0;100" in str(d) for d in called["calls"])


@pytest.mark.asyncio
async def test_set_transition_time(monkeypatch):
    """Test MyStromBulb.set_transition_time sends correct request and returns response."""
    mac = "AA:BB:CC:DD:EE:FF"
    called = {}

    async def _fake_request(
        self, uri, method="POST", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to capture call parameters."""
        called.update({"uri": uri, "method": method, "data": data, "token": token})
        return {"result": "ok"}

    monkeypatch.setattr(bulb_module, "request", _fake_request)
    bulb = MyStromBulb("127.0.0.1", mac)
    resp = await bulb.set_transition_time(123)
    assert called["data"]["ramp"] == 123
    assert resp == {"result": "ok"}


@pytest.mark.asyncio
async def test_context_manager(monkeypatch):
    """Test MyStromBulb can be used as an async context manager."""
    mac = "AA:BB:CC:DD:EE:FF"
    closed = {"called": False}

    async def _fake_close(self):
        """Fake close function to capture call."""
        closed["called"] = True

    monkeypatch.setattr(MyStromBulb, "close", _fake_close)
    async with MyStromBulb("127.0.0.1", mac) as bulb:
        assert isinstance(bulb, MyStromBulb)
    # __aexit__ should call close
    assert closed["called"] is True


@pytest.mark.asyncio
async def test_get_state(monkeypatch):
    """Test MyStromBulb.get_state parses response correctly."""
    mac = "AA:BB:CC:DD:EE:FF"
    fake_response = {
        mac: {
            "power": 5.5,
            "fw_version": "1.2.3",
            "color": "FF000000",
            "mode": "white",
            "ramp": 42,
            "on": 1,
            "type": "RGB",
        }
    }

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to return predefined response."""
        return fake_response

    monkeypatch.setattr(bulb_module, "request", _fake_request)
    bulb = MyStromBulb("127.0.0.1", mac)
    await bulb.get_state()
    assert bulb.mac == mac
    assert bulb.firmware == "1.2.3"
    assert bulb.color == "FF000000"
    assert bulb.mode == "white"
    assert bulb.transition_time == 42
    assert bulb.state is True
    assert bulb.bulb_type == "RGB"
    assert bulb.consumption == 5.5



"""Example code for getting the data from a myStrom bulb."""

from pymystrom import bulb

bulb = bulb.MyStromBulb("192.168.0.51", "5CCF7FA0AFB0")

# Get the details of the bulb
print("Bulb state:", bulb.get_bulb_state())
print("Current color:", bulb.get_color())
print("Brightness:", bulb.get_brightness())
print("Power consumption:", bulb.get_power())
print("Transition time:", bulb.get_transition_time())
print("Firmware version:", bulb.get_firmware())



"""Tests for myStrom Switch devices."""

import pytest

import pymystrom.switch as switch_module
from pymystrom.switch import MyStromSwitch


@pytest.mark.asyncio
async def test_turn_on(monkeypatch):
    """Test MyStromSwitch.turn_on sends correct request and updates state."""
    called = {}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to capture call parameters."""
        called["uri"] = str(uri)
        called["params"] = params
        return {"relay": True, "power": 1.0, "Ws": 0.5}

    monkeypatch.setattr(switch_module, "request", _fake_request)
    sw = MyStromSwitch("127.0.0.1")

    # Patch get_state to avoid recursion
    async def fake_get_state():
        sw._state = True

    sw.get_state = fake_get_state
    await sw.turn_on()
    assert called["params"] == {"state": "1"}


@pytest.mark.asyncio
async def test_turn_off(monkeypatch):
    """Test MyStromSwitch.turn_off sends correct request and updates state."""
    called = {}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to capture call parameters."""
        called["uri"] = str(uri)
        called["params"] = params
        return {"relay": False, "power": 0.0, "Ws": 0.0}

    monkeypatch.setattr(switch_module, "request", _fake_request)
    sw = MyStromSwitch("127.0.0.1")

    async def fake_get_state():
        """Fake get_state to set initial state."""
        sw._state = False

    sw.get_state = fake_get_state
    await sw.turn_off()
    assert called["params"] == {"state": "0"}


@pytest.mark.asyncio
async def test_toggle(monkeypatch):
    """Test MyStromSwitch.toggle sends correct request based on current state."""
    called = {}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to capture call parameters."""
        called["uri"] = str(uri)
        return {"relay": True, "power": 1.0, "Ws": 0.5}

    monkeypatch.setattr(switch_module, "request", _fake_request)
    sw = MyStromSwitch("127.0.0.1")

    async def fake_get_state():
        """Fake get_state to set initial state."""
        sw._state = True

    sw.get_state = fake_get_state
    await sw.toggle()
    assert "toggle" in called["uri"]


@pytest.mark.asyncio
async def test_get_state(monkeypatch):
    """Test MyStromSwitch.get_state sends correct request and updates state."""
    mac = "AA:BB:CC:DD:EE:FF"
    fake_report = {"relay": True, "power": 1.23, "Ws": 0.5}
    fake_info = {"version": "2.68.10", "mac": mac, "type": "SWITCH"}
    responses = [fake_report, fake_info]

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to return predefined responses in sequence."""
        return responses.pop(0)

    monkeypatch.setattr(switch_module, "request", _fake_request)
    sw = MyStromSwitch("127.0.0.1")
    await sw.get_state()
    assert sw.relay is True
    assert sw.consumption == 1.2
    assert sw.consumedWs == 0.5
    assert sw.firmware == "2.68.10"
    assert sw.mac == mac
    assert (
        sw.device_type is None
    )  # Because DEVICE_MAPPING_LITERAL.get("SWITCH") is None by default



"""Example code for communicating with a myStrom PIR unit."""

import asyncio

from pymystrom.pir import MyStromPir

IP_ADDRESS = "192.168.1.225"


async def main():
    """Sample code to work with a myStrom PIR."""
    async with MyStromPir(IP_ADDRESS) as pir:
        # Get the PIR settings
        await pir.get_settings()
        print("Settings:", pir.settings)

        # Get the PIR settings
        await pir.get_pir()
        print("PIR settings:", pir.pir)

        # Collect the sensors
        await pir.get_sensors_state()
        print("Sensors:", pir.sensors)

        # Get the temperature data
        await pir.get_temperatures()
        print("Temperatures:", pir.temperature_raw)
        print("Temperature measured:", pir.temperature_measured)

        # Details of the light sensor
        await pir.get_light()
        print("Brightness:", pir.intensity)
        print("Day?:", pir.day)
        print("Raw light data:", pir.light_raw)

        # Get the action settings
        await pir.get_actions()
        print("Actions:", pir.actions)


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



"""All exceptions for the myStrom Python bindings."""


class MyStromError(Exception):
    """General MyStromError exception occurred."""

    pass


class MyStromConnectionError(MyStromError):
    """When a connection error is encountered."""

    pass


class MyStromNotVersionTwoSwitch(MyStromError):
    """When version 2 function is not supported."""

    pass



"""Tests for myStrom PIR sensors."""

import pytest

import pymystrom.pir as pir_module
from pymystrom.pir import MyStromPir


@pytest.mark.asyncio
async def test_get_settings(monkeypatch):
    """Test MyStromPir.get_settings sends correct request and updates settings."""
    fake_response = {"foo": "bar"}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to return predefined response."""
        return fake_response

    monkeypatch.setattr(pir_module, "request", _fake_request)
    pir = MyStromPir("127.0.0.1")
    await pir.get_settings()
    assert pir.settings == fake_response


@pytest.mark.asyncio
async def test_get_actions(monkeypatch):
    """Test MyStromPir.get_actions sends correct request and updates actions."""
    fake_response = {"action": "test"}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to return predefined response."""
        return fake_response

    monkeypatch.setattr(pir_module, "request", _fake_request)
    pir = MyStromPir("127.0.0.1")
    await pir.get_actions()
    assert pir.actions == fake_response


@pytest.mark.asyncio
async def test_get_pir(monkeypatch):
    """Test MyStromPir.get_pir sends correct request and updates pir settings."""
    fake_response = {"pir": 1}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to return predefined response."""

        return fake_response

    monkeypatch.setattr(pir_module, "request", _fake_request)
    pir = MyStromPir("127.0.0.1")
    await pir.get_pir()
    assert pir.pir == fake_response


@pytest.mark.asyncio
async def test_get_sensors_state(monkeypatch):
    """Test MyStromPir.get_sensors_state sends correct request and updates sensor values."""
    fake_response = {"motion": True, "light": 42, "temperature": 21.2345}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to return predefined response."""
        return fake_response

    monkeypatch.setattr(pir_module, "request", _fake_request)
    pir = MyStromPir("127.0.0.1")
    await pir.get_sensors_state()
    assert pir.sensors == {"motion": True, "light": 42, "temperature": 21.23}


@pytest.mark.asyncio
async def test_get_temperatures(monkeypatch):
    """Test MyStromPir.get_temperatures sends correct request and updates temperature values."""
    fake_response = {"measured": 22.345, "compensated": 23.456, "compensation": 0.12345}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to return predefined response."""
        return fake_response

    monkeypatch.setattr(pir_module, "request", _fake_request)
    pir = MyStromPir("127.0.0.1")
    await pir.get_temperatures()
    assert pir._temperature_raw == fake_response
    assert pir.temperature_measured == 22.34
    assert pir._temperature_compensated == 23.46
    assert pir._temperature_compensation == 0.123


@pytest.mark.asyncio
async def test_get_motion(monkeypatch):
    """Test MyStromPir.get_motion sends correct request and updates motion state."""
    fake_response = {"motion": 1}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to return predefined response."""

        return fake_response

    monkeypatch.setattr(pir_module, "request", _fake_request)
    pir = MyStromPir("127.0.0.1")
    await pir.get_motion()
    assert pir._motion == 1


@pytest.mark.asyncio
async def test_get_light(monkeypatch):
    """Test MyStromPir.get_light sends correct request and updates light sensor values."""
    fake_response = {"intensity": 99, "day": True, "raw": 123}

    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        """Fake request function to return predefined response."""
        return fake_response

    monkeypatch.setattr(pir_module, "request", _fake_request)
    pir = MyStromPir("127.0.0.1")
    await pir.get_light()
    assert pir._intensity == 99
    assert pir._day is True
    assert pir._light_raw == 123



import asyncio

from pymystrom import discovery


async def main():
    """Sample code to work with discovery."""
    # Discover all bulbs in the network via broadcast datagram (UDP)
    bulbs = await discovery.discover_devices()
    print(f"Number of detected bulbs: {len(bulbs)}")


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



"""Test handling of missing 'type' field in device info."""

import pytest

import pymystrom.switch as switch_module
from pymystrom.switch import MyStromSwitch


@pytest.mark.asyncio
async def test_get_state_missing_type(monkeypatch):
    async def _fake_request(
        self, uri, method="GET", data=None, json_data=None, params=None, token=None
    ):
        uri_str = str(uri)
        if uri_str.endswith("/report"):
            return {"relay": True, "power": 1.23, "Ws": 0.5}
        if uri_str.endswith("/api/v1/info"):
            # Legacy v1 firmware without 'type'
            return {"version": "2.68.10", "mac": "AA:BB:CC:DD:EE:FF"}
        if uri_str.endswith("/info.json"):
            return {"version": "2.68.10", "mac": "AA:BB:CC:DD:EE:FF"}
        return {}

    monkeypatch.setattr(switch_module, "request", _fake_request)
    sw = MyStromSwitch("127.0.0.1")
    await sw.get_state()

    assert sw.relay is True
    assert sw.consumption == 1.2
    assert sw.consumedWs == 0.5
    assert sw.firmware == "2.68.10"
    assert sw.mac == "AA:BB:CC:DD:EE:FF"
    # Missing 'type' should be tolerated and map to None
    assert sw.device_type is None



"""Example code for communicating with a myStrom bulb and HSV values."""

import time

from pymystrom import bulb

bulb = bulb.MyStromBulb("192.168.0.51", "5CCF7FA0AFB0")

bulb.set_color_hex("000000FF")

# Get the details of the bulb
print("Current color details:", bulb.get_color())

# Set color as HSV (Hue, Saturation, Value)
bulb.set_color_hsv(50, 100, 100)
time.sleep(3)

# Shutdown the bulb
bulb.set_off()



"""Support for communicating with myStrom plugs/switches."""

from typing import Optional, Union

import aiohttp
from yarl import URL

from . import _request as request
from .device_types import DEVICE_MAPPING_LITERAL, DEVICE_MAPPING_NUMERIC


class MyStromSwitch:
    """A class for a myStrom switch/plug."""

    def __init__(
        self,
        host: str,
        session: aiohttp.client.ClientSession = None,
        token: Optional[str] = None,
    ) -> None:
        """Initialize the switch."""
        self._close_session = False
        self._host = host
        self._token = token
        self._session = session
        self._consumption = 0
        self._consumedWs = 0
        self._boot_id = None
        self._energy_since_boot = None
        self._time_since_boot = None
        self._state = None
        self._temperature = None
        self._firmware = None
        self._mac = None
        self._device_type: Optional[Union[str, int]] = None
        self.uri = URL.build(scheme="http", host=self._host)

    async def turn_on(self) -> None:
        """Turn the relay on."""
        parameters = {"state": "1"}
        url = URL(self.uri).join(URL("relay"))
        await request(self, uri=url, params=parameters, token=self._token)
        await self.get_state()

    async def turn_off(self) -> None:
        """Turn the relay off."""
        parameters = {"state": "0"}
        url = URL(self.uri).join(URL("relay"))
        await request(self, uri=url, params=parameters, token=self._token)
        await self.get_state()

    async def toggle(self) -> None:
        """Toggle the relay."""
        url = URL(self.uri).join(URL("toggle"))
        await request(self, uri=url, token=self._token)
        await self.get_state()

    async def get_state(self) -> None:
        """Get the details from the switch/plug."""
        url = URL(self.uri).join(URL("report"))
        response = await request(self, uri=url, token=self._token)
        try:
            self._consumption = response["power"]
        except KeyError:
            self._consumption = None
        try:
            self._consumedWs = response["Ws"]
        except KeyError:
            self._consumedWs = None
        try:
            self._boot_id = response["boot_id"]
        except KeyError:
            self._boot_id = None
        try:
            self._energy_since_boot = response["energy_since_boot"]
        except KeyError:
            self._energy_since_boot = None
        try:
            self._time_since_boot = response["time_since_boot"]
        except KeyError:
            self._time_since_boot = None
        self._state = response["relay"]
        try:
            self._temperature = response["temperature"]
        except KeyError:
            self._temperature = None

        # Try the new API (Devices with newer firmware)
        url = URL(self.uri).join(URL("api/v1/info"))
        response = await request(self, uri=url, token=self._token)
        if not isinstance(response, dict):
            # Fall back to the old API version if the device runs with old firmware
            url = URL(self.uri).join(URL("info.json"))
            response = await request(self, uri=url, token=self._token)

        # Tolerate missing keys on legacy firmware (e.g., v1 devices)
        self._firmware = response.get("version")
        self._mac = response.get("mac")
        self._device_type = response.get("type")

    @property
    def device_type(self) -> Optional[str]:
        """Return the device type as string (e.g. "Switch CH v1" or "Button+")."""
        if isinstance(self._device_type, int):
            return DEVICE_MAPPING_NUMERIC.get(self._device_type)
        elif isinstance(self._device_type, str):
            return DEVICE_MAPPING_LITERAL.get(self._device_type)
        return None

    @property
    def relay(self) -> bool:
        """Return the relay state."""
        return bool(self._state)

    @property
    def consumption(self) -> Optional[float]:
        """Return the current power consumption in mWh."""
        if self._consumption is not None:
            return round(self._consumption, 1)

        return self._consumption

    @property
    def consumedWs(self) -> Optional[float]:
        """The average of energy consumed per second since last report call."""
        if self._consumedWs is not None:
            return round(self._consumedWs, 1)

        return self._consumedWs

    @property
    def boot_id(self) -> Optional[str]:
        """A unique identifier to distinguish whether the energy counter has been reset."""
        return self._boot_id

    @property
    def energy_since_boot(self) -> Optional[float]:
        """The total energy in watt seconds (Ws) that has been measured since the last power-up or restart of the device."""
        if self._energy_since_boot is not None:
            return round(self._energy_since_boot, 2)

        return self._energy_since_boot

    @property
    def time_since_boot(self) -> Optional[int]:
        """The time in seconds that has elapsed since the last start or restart of the device."""
        return self._time_since_boot

    @property
    def firmware(self) -> Optional[str]:
        """Return the current firmware."""
        return self._firmware

    @property
    def mac(self) -> Optional[str]:
        """Return the MAC address."""
        return self._mac

    @property
    def temperature(self) -> Optional[float]:
        """Return the current temperature in Celsius."""
        if self._temperature is not None:
            return round(self._temperature, 1)

        return self._temperature

    async def get_temperature_full(self) -> str:
        """Get current temperature in celsius."""
        url = URL(self.uri).join(URL("temp"))
        response = await request(self, uri=url, token=self._token)
        return response

    async def close(self) -> None:
        """Close an open client session."""
        if self._session and self._close_session:
            await self._session.close()

    async def __aenter__(self) -> "MyStromSwitch":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self.close()



"""Support for communicating with myStrom PIRs."""

from typing import Optional

import aiohttp
from yarl import URL

from . import _request as request

URI_PIR = URL("api/v1/")


class MyStromPir:
    """A class for a myStrom PIR."""

    def __init__(
        self,
        host: str,
        session: aiohttp.client.ClientSession = None,
        token: Optional[str] = None,
    ) -> None:
        """Initialize the switch."""
        self._close_session = False
        self._host = host
        self._token = token
        self._session = session
        self._intensity = None
        self._day = None
        self._light_raw = None
        self._sensors = None
        self._temperature_measured = None
        self._temperature_compensated = None
        self._temperature_compensation = None
        self._temperature_raw = None
        self._motion = None
        self._settings = None
        self._pir = None

        self._actions = None
        self.uri = URL.build(scheme="http", host=self._host).join(URI_PIR)

    async def get_settings(self) -> None:
        """Get the current settings from the PIR."""
        url = URL(self.uri).join(URL("settings"))
        response = await request(self, uri=url, token=self._token)
        self._settings = response

    async def get_actions(self) -> None:
        """Get the current action settings from the PIR."""
        url = URL(self.uri).join(URL("action"))
        response = await request(self, uri=url, token=self._token)
        self._actions = response

    async def get_pir(self) -> None:
        """Get the current PIR settings."""
        url = URL(self.uri).join(URL("settings/pir"))
        response = await request(self, uri=url, token=self._token)
        self._pir = response

    async def get_sensors_state(self) -> None:
        """Get the state of the sensors from the PIR."""
        url = URL(self.uri).join(URL("sensors"))
        response = await request(self, uri=url, token=self._token)
        # The return data has the be re-written as the temperature is not rounded
        self._sensors = {
            "motion": response["motion"],
            "light": response["light"],
            "temperature": round(response["temperature"], 2),
        }

    async def get_temperatures(self) -> None:
        """Get the temperatures from the PIR."""
        # There is a different URL for the temp endpoint
        url = URL.build(scheme="http", host=self._host) / "temp"
        response = await request(self, uri=url, token=self._token)
        self._temperature_raw = response
        self._temperature_measured = round(response["measured"], 2)
        self._temperature_compensated = round(response["compensated"], 2)
        self._temperature_compensation = round(response["compensation"], 3)

    async def get_motion(self) -> None:
        """Get the state of the motion sensor from the PIR."""
        url = URL(self.uri).join(URL("motion"))
        response = await request(self, uri=url, token=self._token)
        self._motion = response["motion"]

    async def get_light(self) -> None:
        """Get the state of the light sensor from the PIR."""
        url = URL(self.uri).join(URL("light"))
        response = await request(self, uri=url, token=self._token)
        self._intensity = response["intensity"]
        self._day = response["day"]
        self._light_raw = response["raw"]

    @property
    def settings(self) -> Optional[dict]:
        """Return current settings."""
        return self._settings

    @property
    def actions(self) -> Optional[dict]:
        """Return current action settings."""
        return self._actions

    @property
    def pir(self) -> Optional[dict]:
        """Return current PIR settings."""
        return self._pir

    @property
    def sensors(self) -> Optional[dict]:
        """Return current sensor values."""
        return self._sensors

    @property
    def temperature_measured(self) -> Optional[str]:
        """Return current measured temperature."""
        return self._temperature_measured

    @property
    def temperature_compensated(self) -> Optional[str]:
        """Return current compensated temperature."""
        return self._temperature_compensated

    @property
    def temperature_compensation(self) -> Optional[str]:
        """Return current temperature compensation."""
        return self._temperature_compensation

    @property
    def temperature_raw(self) -> Optional[dict]:
        """Return current raw temperature values."""
        return self._temperature_raw

    @property
    def motion(self) -> Optional[str]:
        """Return the state of the motion sensor."""
        return self._motion

    @property
    def intensity(self) -> Optional[str]:
        """Return the intensity reported by the light sensor."""
        return self._intensity

    @property
    def day(self) -> Optional[str]:
        """Return the information based on the thresholds set."""
        return self._day

    @property
    def light_raw(self) -> Optional[str]:
        """Return the raw data from the ADC."""
        return {
            "visible": self._light_raw["adc0"],
            "infrared": self._light_raw["adc1"],
        }

    async def close(self) -> None:
        """Close an open client session."""
        if self._session and self._close_session:
            await self._session.close()

    async def __aenter__(self) -> "MyStromPir":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self.close()



"""Command-line tool for working with myStrom devices."""

import click
import requests
import asyncio

from functools import wraps

from pymystrom.bulb import MyStromBulb

URI = "api/v1/device"
TIMEOUT = 5


def coro(f):
    """Allow to use async in click."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        """Async wrapper."""
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.group()
@click.version_option()
def main():
    """Simple command-line tool to get and set the values of myStrom devices.

    This tool can also set the targets of a myStrom button for the different
    available actions single, double, long and touch.
    """


@main.group("config")
def config():
    """Get and set the configuration of a myStrom device."""


@config.command("read")
@click.option(
    "--ip", prompt="IP address of the device", help="IP address of the device."
)
@click.option(
    "--mac",
    prompt="MAC address of the device",
    help="MAC address of the device.",
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
def read_config(ip, mac, token):
    """Read the current configuration of a myStrom device."""
    click.echo("Read configuration from %s" % ip)
    try:
        headers = {"Token": token} if token else {}
        request = requests.get(
            "http://{}/{}/{}/".format(ip, URI, mac), timeout=TIMEOUT, headers=headers
        )
        click.echo(request.json())
    except requests.exceptions.ConnectionError:
        click.echo("Communication issue with the device")


@main.group("button")
def button():
    """Get and set details of a myStrom button."""


@button.command("generic")
@click.option(
    "--ip", prompt="IP address of the button", help="IP address of the button."
)
@click.option(
    "--mac",
    prompt="MAC address of the button",
    help="MAC address of the button.",
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
@click.option(
    "--single",
    prompt="URL for a single tap",
    default="",
    help="URL for a single tap.",
)
@click.option(
    "--double",
    prompt="URL for a double tap",
    default="",
    help="URL for a double tap.",
)
@click.option(
    "--long", prompt="URL for a long tab", default="", help="URL for a long tab."
)
@click.option("--touch", prompt="URL for a touch", default="", help="URL for a touch.")
def write_config(ip, mac, token, single, double, long, touch):
    """Write the current configuration of a myStrom button."""
    click.echo("Write configuration to device %s" % ip)
    data = {
        "single": single,
        "double": double,
        "long": long,
        "touch": touch,
    }
    try:
        headers = {"Token": token} if token else {}
        request = requests.post(
            "http://{}/{}/{}/".format(ip, URI, mac),
            data=data,
            timeout=TIMEOUT,
            headers=headers,
        )

        if request.status_code == 200:
            click.echo("Configuration of %s set" % mac)
    except requests.exceptions.ConnectionError:
        click.echo("Communication issue with the device. No action performed")


@button.command("home-assistant")
@click.option(
    "--ip", prompt="IP address of the button", help="IP address of the button."
)
@click.option(
    "--mac",
    prompt="MAC address of the button",
    help="MAC address of the button.",
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
@click.option(
    "--hass",
    prompt="IP address of the Home Assistant instance",
    help="IP address of Home Assistant instance to use.",
)
@click.option(
    "--port",
    prompt="Port of Home Assistant instance",
    default="8123",
    help="Port where Home Assistant instance is listening.",
)
@click.option(
    "--id",
    prompt="ID of the button",
    default="",
    help="ID of the myStrom button.",
)
def write_ha_config(ip, mac, token, hass, port, id):
    """Write the configuration for Home Assistant to a myStrom button."""
    click.echo("Write configuration for Home Assistant to device %s..." % ip)

    action = "get://{1}:{2}/api/mystrom?{0}={3}"
    data = {
        "single": action.format("single", hass, port, id),
        "double": action.format("double", hass, port, id),
        "long": action.format("long", hass, port, id),
        "touch": action.format("touch", hass, port, id),
    }
    try:
        headers = {"Token": token} if token else {}
        request = requests.post(
            "http://{}/{}/{}/".format(ip, URI, mac),
            data=data,
            timeout=TIMEOUT,
            headers=headers,
        )

        if request.status_code == 200:
            click.echo("Configuration for %s set" % ip)
            click.echo(
                "After using the push pattern the first time then "
                "the myStrom WiFi Button will show up as %s" % id
            )
    except requests.exceptions.ConnectionError:
        click.echo("Communication issue with the device")


@button.command("reset")
@click.option(
    "--ip",
    prompt="IP address of the WiFi Button",
    help="P address of the WiFi Button.",
)
@click.option(
    "--mac",
    prompt="MAC address of the button",
    help="MAC address of the Wifi Button.",
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
def reset_config(ip, mac, token):
    """Reset the current configuration of a myStrom WiFi Button."""
    click.echo("Reset configuration of button %s..." % ip)
    data = {
        "single": "",
        "double": "",
        "long": "",
        "touch": "",
    }
    try:
        headers = {"Token": token} if token else {}
        request = requests.post(
            "http://{}/{}/{}/".format(ip, URI, mac),
            data=data,
            timeout=TIMEOUT,
            headers=headers,
        )

        if request.status_code == 200:
            click.echo("Reset configuration of %s" % mac)
    except requests.exceptions.ConnectionError:
        click.echo("Communication issue with the device. No action performed")


@button.command("read")
@click.option(
    "--ip",
    prompt="IP address of the WiFi Button",
    help="P address of the WiFi Button.",
)
@click.option(
    "--mac",
    prompt="MAC address of the button",
    help="MAC address of the Wifi Button.",
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
def read_config_button(ip, mac, token):
    """Read the current configuration of a myStrom WiFi Button."""
    click.echo("Read the configuration of button %s..." % ip)
    try:
        headers = {"Token": token} if token else {}
        request = requests.get(
            "http://{}/{}/{}/".format(ip, URI, mac), timeout=TIMEOUT, headers=headers
        )
        click.echo(request.json())
    except requests.exceptions.ConnectionError:
        click.echo("Communication issue with the device. No action performed")


@main.group("bulb")
def bulb():
    """Get and set details of a myStrom bulb."""


@bulb.command("on")
@coro
@click.option("--ip", prompt="IP address of the bulb", help="IP address of the bulb.")
@click.option(
    "--mac", prompt="MAC address of the bulb", help="MAC address of the bulb."
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
async def on(ip, mac, token):
    """Switch the bulb on."""
    async with MyStromBulb(ip=ip, mac=mac, token=token) as bulb:
        await bulb.set_color_hex("00FFFFFF")


@bulb.command("color")
@coro
@click.option("--ip", prompt="IP address of the bulb", help="IP address of the bulb.")
@click.option(
    "--mac", prompt="MAC address of the bulb", help="MAC address of the bulb."
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
@click.option(
    "--hue", prompt="Set the hue of the bulb", help="Set the hue of the bulb."
)
@click.option(
    "--saturation",
    prompt="Set the saturation of the bulb",
    help="Set the saturation of the bulb.",
)
@click.option(
    "--value",
    prompt="Set the value of the bulb",
    help="Set the value of the bulb.",
)
async def color(ip, mac, token, hue, saturation, value):
    """Switch the bulb on with the given color."""
    async with MyStromBulb(ip=ip, mac=mac, token=token) as bulb:
        await bulb.set_color_hsv(hue, saturation, value)


@bulb.command("off")
@coro
@click.option("--ip", prompt="IP address of the bulb", help="IP address of the bulb.")
@click.option(
    "--mac", prompt="MAC address of the bulb", help="MAC address of the bulb."
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
async def off(ip, mac, token):
    """Switch the bulb off."""
    async with MyStromBulb(ip=ip, mac=mac, token=token) as bulb:
        await bulb.set_off()


@bulb.command("flash")
@coro
@click.option("--ip", prompt="IP address of the bulb", help="IP address of the bulb.")
@click.option(
    "--mac", prompt="MAC address of the bulb", help="MAC address of the bulb."
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
@click.option(
    "--time",
    prompt="Time to flash",
    help="Time to flash the bulb in seconds.",
    default=10,
)
async def flash(ip, mac, token, time):
    """Flash the bulb off."""
    async with MyStromBulb(ip=ip, mac=mac, token=token) as bulb:
        await bulb.set_flashing(time, [100, 50, 30], [200, 0, 71])


@bulb.command("rainbow")
@coro
@click.option("--ip", prompt="IP address of the bulb", help="IP address of the bulb.")
@click.option(
    "--mac", prompt="MAC address of the bulb", help="MAC address of the bulb."
)
@click.option("--token", prompt="Token of the device", help="Token of the device.")
@click.option(
    "--time",
    prompt="Time for the complete rainbow",
    help="Time to perform the rainbow in seconds.",
    default=30,
)
async def rainbow(ip, mac, token, time):
    """Let the buld change the color and show a rainbow."""
    async with MyStromBulb(ip=ip, mac=mac, token=token) as bulb:
        await bulb.set_rainbow(time)
        await bulb.set_transition_time(1000)


if __name__ == "__main__":
    main()



"""Example code for communicating with a myStrom bulb."""

import asyncio
import logging

from pymystrom.bulb import MyStromBulb
from pymystrom.discovery import discover_devices

IP_ADDRESS = "192.168.1.25"
MAC_ADDRESS = "5CCF7FA0C27C"


async def main():
    """Sample code to work with a myStrom bulb."""
    # Discover myStrom bulbs devices
    devices = await discover_devices()

    print(f"Found {len(devices)} bulb(s)")
    for device in devices:
        print(f"  IP address: {device.host}, MAC address: {device.mac}")

    async with MyStromBulb(IP_ADDRESS, MAC_ADDRESS) as bulb:
        print("Get the details from the bulb...")
        await bulb.get_state()

        print("Power consumption:", bulb.consumption)
        print("Firmware:", bulb.firmware)
        print("Current state:", "off" if bulb.state is False else "on")

        print("Bulb will be switched on with their previous setting")
        await bulb.set_on()
        # print("Waiting for a couple of seconds...")
        await asyncio.sleep(2)
        print("Bulb will be set to white")
        await bulb.set_white()
        # Wait a few seconds to get a reading of the power consumption
        print("Waiting for a couple of seconds...")

        await asyncio.sleep(2)
        # Set transition time to 2 s
        await bulb.set_transition_time(2000)

        # Set to blue as HEX
        await bulb.set_color_hex("000000FF")
        await asyncio.sleep(3)

        # Set color as HSV (Hue, Saturation, Value)
        await bulb.set_color_hsv(0, 0, 100)
        await asyncio.sleep(3)

        # Test a fast flashing sequence
        print("Flash it for 10 seconds...")
        await bulb.set_flashing(10, [100, 50, 30], [200, 0, 71])
        await bulb.set_off()

        # Show a sunrise within a minute
        print("Show a sunrise for 60 s")
        await bulb.set_sunrise(60)

        # Show a rainbow for 60 seconds
        print("Show a rainbow")
        await bulb.set_rainbow(60)
        # Reset transition time
        await bulb.set_transition_time(1000)

        # Shutdown the bulb
        await bulb.set_off()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



"""
Device types.

See https://api.mystrom.ch/#f37a4be7-0233-4d93-915e-c6f92656f129
"""

DEVICE_MAPPING_NUMERIC = {
    101: "Switch CH v1",
    102: "Bulb",
    103: "Button+",
    104: "Button",
    105: "LED Strip",
    106: "Switch CH v2",
    107: "Switch EU",
    110: "Motion Sensor",
    113: "modulo® STECCO / CUBO",
    118: "Button Plus 2nd",
    120: "Switch Zero",
}

DEVICE_MAPPING_LITERAL = {
    "WSW": DEVICE_MAPPING_NUMERIC[101],
    "WRB": DEVICE_MAPPING_NUMERIC[102],
    "WBP": DEVICE_MAPPING_NUMERIC[103],
    "WBS": DEVICE_MAPPING_NUMERIC[104],
    "WRS": DEVICE_MAPPING_NUMERIC[105],
    "WS2": DEVICE_MAPPING_NUMERIC[106],
    "WSE": DEVICE_MAPPING_NUMERIC[107],
    "WMS": DEVICE_MAPPING_NUMERIC[110],
    "WLL": DEVICE_MAPPING_NUMERIC[113],
    "BP2": DEVICE_MAPPING_NUMERIC[118],
    "LCS": DEVICE_MAPPING_NUMERIC[120],
}



"""Example code for communicating with a myStrom plug/switch."""

import asyncio

from pymystrom.switch import MyStromSwitch

IP_ADDRESS = "192.168.0.40"


async def main():
    """Sample code to work with a myStrom switch."""
    async with MyStromSwitch(IP_ADDRESS) as switch:
        # Collect the data of the current state
        await switch.get_state()

        print("Device type:", switch.device_type)
        print("Power consumption:", switch.consumption)
        print("Energy consumed:", switch.consumedWs)
        print("Relay state:", switch.relay)
        print("Temperature:", switch.temperature)
        print("Firmware:", switch.firmware)
        print("MAC address:", switch.mac)

        print("Turn on the switch")
        if not switch.relay:
            await switch.turn_on()

        # print("Toggle the switch")
        # await switch.toggle()

        # Switch relay off if it was off
        if switch.relay:
            await switch.turn_off()


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



"""Base details for the myStrom Python bindings."""

import asyncio
import socket
from typing import Any, Mapping, Optional

import aiohttp
from yarl import URL

from .exceptions import MyStromConnectionError

TIMEOUT = 10
USER_AGENT = "PythonMyStrom/1.0"


async def _request(
    self,
    uri: str,
    method: str = "GET",
    data: Optional[Any] = None,
    json_data: Optional[dict] = None,
    params: Optional[Mapping[str, str]] = None,
    token: Optional[str] = None,
) -> Any:
    """Handle a request to the myStrom device."""
    headers = {
        "User-Agent": USER_AGENT,
        "Accept": "application/json, text/plain, */*",
    }
    if token:
        headers["Token"] = token

    if self._session is None:
        self._session = aiohttp.ClientSession()
        self._close_session = True

    try:
        response = await asyncio.wait_for(
            self._session.request(
                method,
                uri,
                data=data,
                json=json_data,
                params=params,
                headers=headers,
            ),
            timeout=TIMEOUT,
        )
    except asyncio.TimeoutError as exception:
        raise MyStromConnectionError(
            "Timeout occurred while connecting to myStrom device."
        ) from exception
    except (aiohttp.ClientError, socket.gaierror) as exception:
        raise MyStromConnectionError(
            "Error occurred while communicating with myStrom device."
        ) from exception

    content_type = response.headers.get("Content-Type", "")
    if response.status == 404:
        raise MyStromConnectionError(
            "Error occurred while authenticating with myStrom device."
        )

    elif (response.status // 100) in [4, 5]:
        response.close()

    if "application/json" in content_type:
        response_json = await response.json()
        return response_json

    return response.text


class MyStromDevice:
    """A class for a myStrom device."""

    def __init__(
        self,
        host,
        session: aiohttp.client.ClientSession = None,
    ):
        """Initialize the device."""
        self._close_session = False
        self._host = host
        self._session = session
        self.uri = URL.build(scheme="http", host=self._host)

    async def get_device_info(self) -> dict:
        """Get the device info of a myStrom device."""
        url = URL(self.uri).join(URL("api/v1/info"))
        response = await _request(self, uri=url)
        if not isinstance(response, dict):
            # Fall back to the old API version if the device runs with old firmware
            url = URL(self.uri).join(URL("info.json"))
            response = await _request(self, uri=url)
        return response

    async def close(self) -> None:
        """Close an open client session."""
        if self._session and self._close_session:
            await self._session.close()

    async def __aenter__(self) -> "MyStromDevice":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self.close()


async def get_device_info(host: str) -> dict:
    """Get the device info of a myStrom device."""
    async with MyStromDevice(host) as device:
        return await device.get_device_info()



"""Support for communicating with myStrom bulbs."""

import asyncio
import logging
from typing import Optional

import aiohttp
from yarl import URL

from . import _request as request

_LOGGER = logging.getLogger(__name__)

URI_BULB = URL("api/v1/device")


class MyStromBulb:
    """A class for a myStrom bulb."""

    def __init__(
        self,
        host: str,
        mac: str,
        token: Optional[str] = None,
        session: aiohttp.client.ClientSession = None,
    ) -> None:
        """Initialize the bulb."""
        self._close_session = False
        self._host = host
        self._mac = mac
        self._session = session
        self.brightness = 0
        self._color = None
        self._consumption = 0
        self.data = None
        self._firmware = None
        self._mode = None
        self._bulb_type = None
        self._state = None
        self._transition_time = 0
        self.uri = URL.build(scheme="http", host=self._host).join(URI_BULB) / self._mac
        self.token = token

    async def get_state(self) -> None:
        """Get the state of the bulb."""
        response = await request(self, uri=self.uri, token=self.token)
        self._consumption = response[self._mac]["power"]
        self._firmware = response[self._mac]["fw_version"]
        self._color = response[self._mac]["color"]
        self._mode = response[self._mac]["mode"]
        self._transition_time = response[self._mac]["ramp"]
        self._state = bool(response[self._mac]["on"])
        self._bulb_type = response[self._mac]["type"]

    @property
    def firmware(self) -> Optional[str]:
        """Return current firmware."""
        return self._firmware

    @property
    def mac(self) -> str:
        """Return the MAC address."""
        return self._mac

    @property
    def consumption(self) -> Optional[float]:
        """Return current firmware."""
        return self._consumption

    @property
    def color(self) -> Optional[str]:
        """Return current color settings."""
        return self._color

    @property
    def mode(self) -> Optional[str]:
        """Return current mode."""
        return self._mode

    @property
    def transition_time(self) -> Optional[int]:
        """Return current transition time (ramp)."""
        return self._transition_time

    @property
    def bulb_type(self) -> Optional[str]:
        """Return the type of the bulb."""
        return self._bulb_type

    @property
    def state(self) -> Optional[str]:
        """Return the current state of the bulb."""
        return self._state

    async def set_on(self):
        """Turn the bulb on with the previous settings."""
        response = await request(
            self, uri=self.uri, method="POST", data={"action": "on"}, token=self.token
        )
        return response

    async def set_color_hex(self, value):
        """Turn the bulb on with the given color as HEX.

        white: FF000000
        red:   00FF0000
        green: 0000FF00
        blue:  000000FF
        """
        data = {
            "action": "on",
            "color": value,
        }
        response = await request(
            self, uri=self.uri, method="POST", data=data, token=self.token
        )
        return response

    async def set_color_hsv(self, hue, saturation, value):
        """Turn the bulb on with the given values as HSV."""
        # The current situation doesn't allow to send JSON to the bulb as
        # the firmware wants a string separated by ;. This is was
        # reported in 2018 to myStrom
        # data = {
        #     'action': 'on',
        #     'color': f"{hue};{saturation};{value}",
        # }
        data = "action=on&color={};{};{}".format(hue, saturation, value)
        response = await request(
            self, uri=self.uri, method="POST", data=data, token=self.token
        )
        return response

    async def set_white(self):
        """Turn the bulb on, full white."""
        await self.set_color_hsv(0, 0, 100)

    async def set_rainbow(self, duration):
        """Turn the bulb on and create a rainbow."""
        for i in range(0, 359):
            await self.set_color_hsv(i, 100, 100)
            await asyncio.sleep(duration / 359)

    async def set_sunrise(self, duration):
        """Turn the bulb on and create a sunrise.

        The brightness is from 0 till 100.
        """
        max_brightness = 100
        await self.set_transition_time((duration / max_brightness))
        for i in range(0, duration):
            data = "action=on&color=3;{}".format(i)
            await request(
                self, uri=self.uri, method="POST", data=data, token=self.token
            )
            await asyncio.sleep(duration / max_brightness)

    async def set_flashing(self, duration, hsv1, hsv2):
        """Turn the bulb on, flashing with two colors."""
        await self.set_transition_time(100)
        for step in range(0, int(duration / 2)):
            await self.set_color_hsv(hsv1[0], hsv1[1], hsv1[2])
            await asyncio.sleep(1)
            await self.set_color_hsv(hsv2[0], hsv2[1], hsv2[2])
            await asyncio.sleep(1)

    async def set_transition_time(self, value):
        """Set the transition time in ms."""
        response = await request(
            self,
            uri=self.uri,
            method="POST",
            data={"ramp": int(round(value))},
            token=self.token,
        )
        return response

    async def set_off(self):
        """Turn the bulb off."""
        response = await request(
            self, uri=self.uri, method="POST", data={"action": "off"}, token=self.token
        )
        return response

    async def close(self) -> None:
        """Close an open client session."""
        if self._session and self._close_session:
            await self._session.close()

    async def __aenter__(self) -> "MyStromBulb":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self.close()
